@@ -15,10 +15,10 @@ To use, declare this program as a process in your
        peer-as 64497;
     }
     process watch-haproxy {
-       run python -m exabgp healthcheck --cmd "curl -sf http://127.0.0.1/healthcheck" --label haproxy;
+       run exabgp-healthcheck --cmd "curl -sf http://127.0.0.1/healthcheck" --label haproxy;
     }
     process watch-mysql {
-       run python -m exabgp healthcheck --cmd "mysql -u check -e 'SELECT 1'" --label mysql;
+       run exabgp-healthcheck --cmd "mysql -u check -e 'SELECT 1'" --label mysql;
     }
 
 Use :option:`--help` to get options accepted by this program. A
@@ -43,7 +43,6 @@ labelled ``lo:haproxy1``, ``lo:haproxy2``, etc.
 from __future__ import print_function
 from __future__ import unicode_literals
 
-import string
 import sys
 import os
 import subprocess
@@ -54,37 +53,60 @@ import argparse
 import signal
 import time
 import collections
-try:
-    # Python 3.3+ or backport
-    from ipaddress import ip_address as _ip_address  # pylint: disable=F0401
-
-    def ip_address(x):
-        try:
-            x = x.decode('ascii')
-        except AttributeError:
-            pass
-        return _ip_address(x)
-except ImportError:
-    # Python 2.6, 2.7, 3.2
-    from ipaddr import IPAddress as ip_address
 
 logger = logging.getLogger("healthcheck")
 
+try:
+    # Python 3.3+ or backport
+    from ipaddress import ip_network as ip_network  # pylint: disable=F0401
+    from ipaddress import ip_address as ip_address  # pylint: disable=F0401
 
-def enum (*sequential):
+    def fix(f):
+        def fixed(x):
+            try:
+                x = x.decode('ascii')
+            except AttributeError:
+                pass
+            return f(x)
+        return fixed
+    ip_network = fix(ip_network)
+    ip_address = fix(ip_address)
+
+except ImportError:
+    try:
+        # Python 2.6, 2.7, 3.2
+        from ipaddr import IPNetwork as ip_network
+        from ipaddr import IPAddress as ip_address
+    except ImportError:
+        sys.stderr.write(
+            '\n'
+            'This program requires the python module ip_address (for python 3.3+) or ipaddr (for python 2.6, 2.7, 3.2)\n'
+            'Please pip install one of them with one of the following command.\n'
+            '> pip install ip_address\n'
+            '> pip install ipaddr\n'
+            '\n'
+        )
+        sys.exit(1)
+
+
+def enum(*sequential):
     """Create a simple enumeration."""
     return type(str("Enum"), (), dict(zip(sequential, sequential)))
 
 
-def parse ():
+def parse():
     """Parse arguments"""
+    formatter = argparse.RawDescriptionHelpFormatter
     parser = argparse.ArgumentParser(description=sys.modules[__name__].__doc__,
-                                     formatter_class=argparse.RawDescriptionHelpFormatter)
+                                     formatter_class=formatter)
 
     g = parser.add_mutually_exclusive_group()
     g.add_argument("--debug", "-d", action="store_true",
                    default=False,
                    help="enable debugging")
+    g.add_argument("--no-ack", "-a", action="store_true",
+                   default=False,
+                   help="set for exabgp 3.4 or 4.x when exabgp.api.ack=false")
     g.add_argument("--silent", "-s", action="store_true",
                    default=False,
                    help="don't log to console")
@@ -92,17 +114,23 @@ def parse ():
                    nargs='?',
                    const="daemon",
                    default="daemon",
-                   help="log to syslog using FACILITY, default FACILITY is daemon")
+                   help=("log to syslog using FACILITY, "
+                         "default FACILITY is daemon"))
+    g.add_argument("--sudo", action="store_true", default=False,
+                   help="use sudo to setup ip addresses")
     g.add_argument("--no-syslog", action="store_true",
                    help="disable syslog logging")
     parser.add_argument("--name", "-n", metavar="NAME",
                         help="name for this healthchecker")
     parser.add_argument("--config", "-F", metavar="FILE", type=open,
                         help="read configuration from a file")
-    parser.add_argument("--pid", "-p", metavar="FILE", type=argparse.FileType('w'),
+    parser.add_argument("--pid", "-p", metavar="FILE",
+                        type=argparse.FileType('w'),
                         help="write PID to the provided file")
-    parser.add_argument("--sudo", action="store_true", default=False,
-                        help="use sudo to setup ip addresses")
+    parser.add_argument("--user", metavar="USER",
+                        help="set user after setting loopback addresses")
+    parser.add_argument("--group", metavar="GROUP",
+                        help="set group after setting loopback addresses")
 
     g = parser.add_argument_group("checking healthiness")
     g.add_argument("--interval", "-i", metavar='N',
@@ -112,7 +140,8 @@ def parse ():
     g.add_argument("--fast-interval", "-f", metavar='N',
                    default=1,
                    type=float, dest="fast",
-                   help="when a state change is about to occur, wait N seconds between each healthcheck")
+                   help=("when a state change is about to occur, "
+                         "wait N seconds between each healthcheck"))
     g.add_argument("--timeout", "-t", metavar='N',
                    default=5,
                    type=int,
@@ -137,14 +166,20 @@ def parse ():
                    type=ip_address,
                    help="self IP address to use as next hop")
     g.add_argument("--ip", metavar='IP',
-                   type=ip_address, dest="ips", action="append",
-                   help="advertise this IP address")
+                   type=ip_network, dest="ips", action="append",
+                   help="advertise this IP address or network (CIDR notation)")
+    g.add_argument("--local-preference", metavar='P',
+                   type=int, default=-1,
+                   help="advertise with local preference P")
+    g.add_argument("--deaggregate-networks",
+                   dest="deaggregate_networks", action="store_true",
+                   help="Deaggregate Networks specified in --ip")
     g.add_argument("--no-ip-setup",
                    action="store_false", dest="ip_setup",
                    help="don't setup missing IP addresses")
     g.add_argument("--dynamic-ip-setup", default=False,
                    action="store_true", dest="ip_dynamic",
-                   help="delete existing loopback ips on state down and "\
+                   help="delete existing loopback ips on state down and "
                         "disabled, then restore loopback when up")
     g.add_argument("--label", default=None,
                    help="use the provided label to match loopback addresses")
@@ -159,13 +194,21 @@ def parse ():
                    help="first IP get the metric M when the service is down")
     g.add_argument("--disabled-metric", metavar='M',
                    type=int, default=500,
-                   help="first IP get the metric M when the service is disabled")
+                   help=("first IP get the metric M "
+                         "when the service is disabled"))
     g.add_argument("--increase", metavar='M',
                    type=int, default=1,
-                   help="for each additional IP address increase metric value by W")
+                   help=("for each additional IP address, "
+                         "increase metric value by M"))
     g.add_argument("--community", metavar="COMMUNITY",
                    type=str, default=None,
                    help="announce IPs with the supplied community")
+    g.add_argument("--extended-community", metavar="EXTENDEDCOMMUNITY",
+                   type=str, default=None,
+                   help="announce IPs with the supplied extended community")
+    g.add_argument("--large-community", metavar="LARGECOMMUNITY",
+                   type=str, default=None,
+                   help="announce IPs with the supplied large community")
     g.add_argument("--disabled-community", metavar="DISABLEDCOMMUNITY",
                    type=str, default=None,
                    help="announce IPs with the supplied community when disabled")
@@ -173,7 +216,8 @@ def parse ():
                    type=str, default=None,
                    help="announce IPs with the supplied as-path")
     g.add_argument("--withdraw-on-down", action="store_true",
-                   help="Instead of increasing the metric on health failure, withdraw the route")
+                   help=("Instead of increasing the metric on health failure, "
+                         "withdraw the route"))
 
     g = parser.add_argument_group("reporting")
     g.add_argument("--execute", metavar='CMD',
@@ -202,7 +246,7 @@ def parse ():
     return options
 
 
-def setup_logging (debug, silent, name, syslog_facility, syslog):
+def setup_logging(debug, silent, name, syslog_facility, syslog):
     """Setup logger"""
 
     def syslog_address():
@@ -213,6 +257,8 @@ def setup_logging (debug, silent, name, syslog_facility, syslog):
             return "/var/run/log"
         if sys.platform.startswith("netbsd"):
             return "/var/run/log"
+        if sys.platform.startswith("openbsd"):
+            return "/dev/log"
         if sys.platform.startswith("linux"):
             return "/dev/log"
         raise EnvironmentError("Unable to guess syslog address for your "
@@ -223,7 +269,7 @@ def setup_logging (debug, silent, name, syslog_facility, syslog):
     # To syslog
     if enable_syslog:
         facility = getattr(logging.handlers.SysLogHandler,
-                           "LOG_{0}".format(string.upper(syslog_facility)))
+                           "LOG_{0}".format(syslog_facility.upper()))
         sh = logging.handlers.SysLogHandler(address=str(syslog_address()),
                                             facility=facility)
         if name:
@@ -236,50 +282,66 @@ def setup_logging (debug, silent, name, syslog_facility, syslog):
                 os.getpid())))
         logger.addHandler(sh)
     # To console
-    if sys.stderr.isatty() and not silent:
+    toconsole = (hasattr(sys.stderr, "isatty") and
+                 sys.stderr.isatty() and  # pylint: disable=E1101
+                 not silent)
+    if toconsole:
         ch = logging.StreamHandler()
         ch.setFormatter(logging.Formatter(
             "%(levelname)s[%(name)s] %(message)s"))
         logger.addHandler(ch)
 
 
-def loopback_ips (label):
+def loopback_ips(label):
     """Retrieve loopback IP addresses"""
     logger.debug("Retrieve loopback IP addresses")
     addresses = []
 
     if sys.platform.startswith("linux"):
         # Use "ip" (ifconfig is not able to see all addresses)
-        ipre = re.compile(r"^(?P<index>\d+):\s+(?P<name>\S+)\s+inet6?\s+(?P<ip>[\da-f.:]+)/(?P<netmask>\d+)\s+.*")
-        labelre = re.compile(r".*\s+lo:(?P<label>\S+)\s+.*")
-        cmd = subprocess.Popen("/sbin/ip -o address show dev lo".split(), shell=False, stdout=subprocess.PIPE)
+        ipre = re.compile(r"^(?P<index>\d+):\s+(?P<name>\S+)\s+inet6?\s+"
+                          r"(?P<ip>[\da-f.:]+)/(?P<mask>\d+)\s+.*")
+        labelre = re.compile(r".*\s+lo:(?P<label>\S+).*")
+        cmd = subprocess.Popen("/sbin/ip -o address show dev lo".split(),
+                               shell=False, stdout=subprocess.PIPE)
     else:
         # Try with ifconfig
-        ipre = re.compile(r"^inet6?\s+(alias\s+)?(?P<ip>[\da-f.:]+)\s+(?:netmask 0x(?P<netmask>[0-9a-f]+)|prefixlen (?P<mask>\d+)).*")
-        cmd = subprocess.Popen("/sbin/ifconfig lo0".split(), shell=False, stdout=subprocess.PIPE)
+        ipre = re.compile(r"^inet6?\s+(alias\s+)?(?P<ip>[\da-f.:]+)\s+"
+                          r"(?:netmask 0x(?P<netmask>[0-9a-f]+)|"
+                          r"prefixlen (?P<mask>\d+)).*")
+        cmd = subprocess.Popen("/sbin/ifconfig lo0".split(), shell=False,
+                               stdout=subprocess.PIPE)
         labelre = re.compile(r"")
     for line in cmd.stdout:
         line = line.decode("ascii", "ignore").strip()
         mo = ipre.match(line)
         if not mo:
             continue
-        ip = ip_address(mo.group("ip"))
+        if mo.group("mask"):
+            mask = int(mo.group("mask"))
+        else:
+            mask = bin(int(mo.group("netmask"), 16)).count("1")
+        try:
+            ip = ip_network("{0}/{1}".format(mo.group("ip"),
+                                             mask))
+        except ValueError:
+            continue
         if not ip.is_loopback:
             if label:
                 lmo = labelre.match(line)
                 if not lmo or not lmo.group("label").startswith(label):
                     continue
             addresses.append(ip)
-    logger.debug("Loopback addresses: {0}".format(addresses))
+    logger.debug("Loopback addresses: %s", addresses)
     return addresses
 
 
-def setup_ips (ips, label, sudo=False):
+def setup_ips(ips, label, sudo=False):
     """Setup missing IP on loopback interface"""
     existing = set(loopback_ips(label))
-    toadd = set(ips) - existing
+    toadd = set([ip_network(ip) for net in ips for ip in net]) - existing
     for ip in toadd:
-        logger.debug("Setup loopback IP address {0}".format(ip))
+        logger.debug("Setup loopback IP address %s", ip)
         with open(os.devnull, "w") as fnull:
             cmd = ["ip", "address", "add", str(ip), "dev", "lo"]
             if sudo:
@@ -289,29 +351,57 @@ def setup_ips (ips, label, sudo=False):
             subprocess.check_call(
                 cmd, stdout=fnull, stderr=fnull)
 
-def remove_ips (ips, label, sudo=False):
-    """Remove existing IP on loopback interface"""
+    # If we setup IPs we should also remove them on SIGTERM
+    def sigterm_handler(signum, frame):  # pylint: disable=W0612,W0613
+        remove_ips(ips, label, sudo)
+        sys.exit(0)
+
+    signal.signal(signal.SIGTERM, sigterm_handler)
+
+
+def remove_ips(ips, label, sudo=False):
+    """Remove added IP on loopback interface"""
     existing = set(loopback_ips(label))
 
     # Get intersection of IPs (ips setup, and IPs configured by ExaBGP)
-    toremove = existing.intersection(ips)
+    toremove = set([ip_network(ip) for net in ips for ip in net]) | existing
     for ip in toremove:
-        logger.debug("Remove loopback IP address {0}".format(ip))
+        logger.debug("Remove loopback IP address %s", ip)
         with open(os.devnull, "w") as fnull:
             cmd = ["ip", "address", "delete", str(ip), "dev", "lo"]
             if sudo:
                 cmd.insert(0, "sudo")
             if label:
                 cmd += ["label", "lo:{0}".format(label)]
-            subprocess.check_call(
-                cmd, stdout=fnull, stderr=fnull)
+            try:
+                subprocess.check_call(
+                    cmd, stdout=fnull, stderr=fnull)
+            except subprocess.CalledProcessError:
+                logger.warn("Unable to remove loopback IP address %s - is \
+                    healthcheck running as root?", str(ip))
 
 
-def setpgrp_preexec_fn ():
-    os.setpgrp()
+def drop_privileges(user, group):
+    """Drop privileges to specified user and group"""
+    if group is not None:
+        import grp
+        gid = grp.getgrnam(group).gr_gid
+        logger.debug("Dropping privileges to group {0}/{1}".format(group, gid))
+        try:
+            os.setresgid(gid, gid, gid)
+        except AttributeError:
+            os.setregid(gid, gid)
+    if user is not None:
+        import pwd
+        uid = pwd.getpwnam(user).pw_uid
+        logger.debug("Dropping privileges to user {0}/{1}".format(user, uid))
+        try:
+            os.setresuid(uid, uid, uid)
+        except AttributeError:
+            os.setreuid(uid, uid)
 
 
-def check (cmd, timeout):
+def check(cmd, timeout):
     """Check the return code of the given command.
 
     :param cmd: command to execute. If :keyword:`None`, no command is executed.
@@ -323,16 +413,18 @@ def check (cmd, timeout):
         return True
 
     class Alarm(Exception):
+        """Exception to signal an alarm condition."""
         pass
 
-    def alarm_handler (number, frame):  # pylint: disable=W0613
+    def alarm_handler(number, frame):  # pylint: disable=W0613
+        """Handle SIGALRM signal."""
         raise Alarm()
 
-    logger.debug("Checking command {0}".format(repr(cmd)))
+    logger.debug("Checking command %s", repr(cmd))
     p = subprocess.Popen(cmd, shell=True,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.STDOUT,
-                         preexec_fn=setpgrp_preexec_fn)
+                         preexec_fn=os.setpgrp)
     if timeout:
         signal.signal(signal.SIGALRM, alarm_handler)
         signal.alarm(timeout)
@@ -342,19 +434,22 @@ def check (cmd, timeout):
         if timeout:
             signal.alarm(0)
         if p.returncode != 0:
-            logger.warn("Check command was unsuccessful: {0}".format(p.returncode))
+            logger.warn("Check command was unsuccessful: %s",
+                        p.returncode)
             if stdout.strip():
-                logger.info("Output of check command: {0}".format(stdout))
+                logger.info("Output of check command: %s", stdout)
             return False
-        logger.debug("Command was executed successfully {0} {1}".format(p.returncode, stdout))
+        logger.debug(
+            "Command was executed successfully %s %s", p.returncode, stdout)
         return True
     except Alarm:
-        logger.warn("Timeout ({0}) while running check command {1}".format(timeout, cmd))
+        logger.warn("Timeout (%s) while running check command %s",
+                    timeout, cmd)
         os.killpg(p.pid, signal.SIGKILL)
         return False
 
 
-def loop (options):
+def loop(options):
     """Main loop."""
     states = enum(
         "INIT",                 # Initial state
@@ -364,33 +459,35 @@ def loop (options):
         "UP",                   # Service is considered as up.
         "DOWN",                 # Service is considered as down.
     )
-    state = states.INIT
 
-    def exabgp (target):
+    def exabgp(target):
         """Communicate new state to ExaBGP"""
         if target not in (states.UP, states.DOWN, states.DISABLED):
             return
+        # dynamic ip management. When the service fail, remove the loopback
         if target in (states.DOWN, states.DISABLED) and options.ip_dynamic:
             logger.info("service down, deleting loopback ips")
             remove_ips(options.ips, options.label, options.sudo)
+        # if ips was deleted with dyn ip, re-setup them
         if target == states.UP and options.ip_dynamic:
             logger.info("service up, restoring loopback ips")
             setup_ips(options.ips, options.label, options.sudo)
 
-        logger.info("send announces for {0} state to ExaBGP".format(target))
+        logger.info("send announces for %s state to ExaBGP", target)
         metric = vars(options).get("{0}_metric".format(str(target).lower()))
         for ip in options.ips:
             if options.withdraw_on_down:
                 command = "announce" if target is states.UP else "withdraw"
             else:
                 command = "announce"
-            announce = "route {0}/{1} next-hop {2}".format(
+            announce = "route {0} next-hop {1}".format(
                 str(ip),
-                ip.max_prefixlen,
                 options.next_hop or "self")
+
             if command == "announce":
                 announce = "{0} med {1}".format(announce, metric)
-
+                if options.local_preference >= 0:
+                    announce = "{0} local-preference {1}".format(announce, options.local_preference)
                 if options.community or options.disabled_community:
                     community = options.community
                     if target in (states.DOWN, states.DISABLED):
@@ -399,17 +496,35 @@ def loop (options):
                     if community:
                         announce = "{0} community [ {1} ]".format(
                             announce, community)
-
+                if options.extended_community:
+                    announce = "{0} extended-community [ {1} ]".format(
+                        announce,
+                        options.extended_community)
+                if options.large_community:
+                    announce = "{0} large-community [ {1} ]".format(
+                        announce,
+                        options.large_community)
                 if options.as_path:
                     announce = "{0} as-path [ {1} ]".format(
                         announce,
                         options.as_path)
+
+            metric += options.increase
+
+            # Send and flush command
             logger.debug("exabgp: {0} {1}".format(command, announce))
             print("{0} {1}".format(command, announce))
-            metric += options.increase
-        sys.stdout.flush()
+            sys.stdout.flush()
+            
+            # Wait for confirmation from ExaBGP if expected
+            if options.no_ack:
+                continue
+            # if the program is not ran manually, do not read the input
+            if hasattr(sys.stdout, "isatty") and sys.stdout.isatty():
+                continue
+            sys.stdin.readline()
 
-    def trigger (target):
+    def trigger(target):
         """Trigger a state change and execute the appropriate commands"""
         # Shortcut for RISING->UP and FALLING->UP
         if target == states.RISING and options.rise <= 1:
@@ -418,12 +533,14 @@ def loop (options):
             target = states.DOWN
 
         # Log and execute commands
-        logger.debug("Transition to {0}".format(str(target)))
+        logger.debug("Transition to %s", str(target))
         cmds = []
-        cmds.extend(vars(options).get("{0}_execute".format(str(target).lower()), []) or [])
+        cmds.extend(vars(options).get("{0}_execute".format(
+            str(target).lower()), []) or [])
         cmds.extend(vars(options).get("execute", []) or [])
         for cmd in cmds:
-            logger.debug("Transition to {0}, execute `{1}`".format(str(target), cmd))
+            logger.debug("Transition to %s, execute `%s`",
+                         str(target), cmd)
             env = os.environ.copy()
             env.update({"STATE": str(target)})
             with open(os.devnull, "w") as fnull:
@@ -432,9 +549,10 @@ def loop (options):
 
         return target
 
-    checks = 0
-    while True:
-        disabled = options.disable is not None and os.path.exists(options.disable)
+    def one(checks, state):
+        """Execute one loop iteration."""
+        disabled = (options.disable is not None and
+                    os.path.exists(options.disable))
         successful = disabled or check(options.command, options.timeout)
         # FSM
         if state != states.DISABLED and disabled:
@@ -481,6 +599,12 @@ def loop (options):
         # Send announces. We announce them on a regular basis in case
         # we lose connection with a peer.
         exabgp(state)
+        return checks, state
+
+    checks = 0
+    state = states.INIT
+    while True:
+        checks, state = one(checks, state)
 
         # How much we should sleep?
         if state in (states.FALLING, states.RISING):
@@ -488,7 +612,9 @@ def loop (options):
         else:
             time.sleep(options.interval)
 
-def main ():
+
+def main():
+    """Entry point."""
     options = parse()
     setup_logging(options.debug, options.silent, options.name,
                   options.syslog_facility, not options.no_syslog)
@@ -499,16 +625,25 @@ def main ():
         # Setup IP to use
         options.ips = options.ips or loopback_ips(options.label)
         if not options.ips:
-            raise RuntimeError("No IP found")
+            logger.error("No IP found")
+            sys.exit(1)
         if options.ip_setup:
             setup_ips(options.ips, options.label, options.sudo)
+        drop_privileges(options.user, options.group)
+
+        # Parse defined networks into a list of IPs for advertisement
+        if options.deaggregate_networks:
+            options.ips = [ip_network(ip) for net in options.ips for ip in net]
+
         options.ips = collections.deque(options.ips)
         options.ips.rotate(-options.start_ip)
         options.ips = list(options.ips)
         # Main loop
         loop(options)
-    except Exception as e:
-        logger.exception("Uncatched exception: %s", e)
+    except Exception as e:  # pylint: disable=W0703
+        logger.exception("Uncaught exception: %s", e)
+        sys.exit(1)
+
 
 if __name__ == "__main__":
     main()
